/* ********************************************************************
 * Copyright (c) 2019 Biological Modeling and Visualization Laboratory,
 * University of Calgary. All rights reserved.
 * ********************************************************************/

#include <lpfgall.h>		//include all required files

/* General constant definitions */

#define PI 		3.14159265359
#define PHI 	0.61803398875

#define DT      1.0
#define EPS     1e-5

/* Simulation phase definitions */

#define Grow			1	// Grow the active ring
#define Displace		2	// Adjust primordia position based on forces
#define Insert			3	// Insert primordia if threshold distance exceeded
#define AdjustStatus 	4	// Adjust status of primordia if a new primordium inserted
#define Wait			5	// Redisplay current frame for visualization purposes

/* Cell attribute definitions */

const int S=0;	// short interval between primordia
const int L=1;	// long interval between  primordia 

const int LEFT = -1;	// interval polarity: younger primordium to the left (clockwise) 
const int RIGHT = 1;	// interval polarity: younger primordium to the right (counterclockwise)

/* Model parameters */

#define ratio		0.5	// relative length of non-growing space upon division
#define ThdRatio	0.997	// reduction of threshold distance per step

/* Useful functions */

const double rate = pow(1+PHI, 1.0/200.0);	// relative radius growth per step

double InInterval(double x)	// Circular value math: take x to interval [0,C]
{
	extern double C; 
	if (x>=C)
		return InInterval(x-C);
	if (x<0)
		return InInterval(x+C);
	return x;
}

/* Module declarations */

module P(double, double,int);	// primordium(position, age, mass)
module I(int, int);			// interval between primordia (state S or L, polarity, force)
module Background();				// display a background image
module ShowNumber();				// display current number of primordia

/* External variable declarations */

int Phase;		// simulation phase
double R;		// current radius of the active ring 
double C;		// current length of the active ring
double Thd;		// current threshold

double sim_time;
int step;		// step counter
int n;			// current number of primordia
int InFibo;		// 1 if the current number of primordia is a Fibonacci number, 0 otherwise.
double FiboSum;		// sum of the ring widths with a Fibonacci number of primordia
double NonFiboSum;	// sum of the ring widths with a non-Fibonacci number of primordia
int WaitCount;	// Wait counter
char str[40];	// string for displaying primordia numbers

ring L-system: 1;	// assume circle topology

consider: P I;	// symbols considered by context-sensitive rules

Axiom: SB Background() EB SB SetColor(2) Sphere(0.075) EB  P(0.01,100,1) I(L,RIGHT);

derivation length: 0; // 2500 13P // 2950 21P // 3330 34P

/* Initialize globalvariables */

double k=0.02266;

Start: 
{
	Phase = Insert; 
	R = PHI*3.5; 	// initial radiums of the active zone
	C=2*PI*R; 		// initial circumference of the active zone 
	Thd = 2*PI*3.5;	// initial threshold distance
	step = 0; 		// simulation step
	n=1; 			// initial number of primordia
	InFibo = 1;		// 1 is a Fibonacci number 
	FiboSum = 0;	// initial ring width sum with a Fibonacci number of primordia
	NonFiboSum = 0; // initial ring width sum with a non-Fibonacci number of primordia
	WaitCount=0;

    sim_time = 0.;
} 

/* Control simulation phase; update statistics */

EndEach: 
{
	switch (Phase) {
	case Grow:
		R = R + R * (rate - 1.) * DT;
		C = 2*PI*R;
		Thd = Thd + Thd * (ThdRatio - 1.) * DT;
		if (InFibo == 1)	
			FiboSum += R*(rate-1.0); 
		else
			NonFiboSum += R*(rate-1.0);
		Phase = Displace;
		break;
	case Displace:
		Phase = Insert ;
		break;
	case Insert:
		Phase = AdjustStatus;
		break;
	case AdjustStatus: {
        sim_time += DT;
        if (++step % int(1./DT) == 0)
		    DisplayFrame();

		if (std::abs(sim_time-175.) < EPS ||
            std::abs(sim_time-268.) < EPS ||
            std::abs(sim_time-290.) < EPS ||
            std::abs(sim_time-439.) < EPS ||
            std::abs(sim_time-520.) < EPS ||
            std::abs(sim_time-620.) < EPS)
        {
			WaitCount = 0;
			Phase = Wait;
		}
		else

			Phase = Grow;

        if (sim_time + EPS > 621) {
			WaitCount = 100;
            Phase = Wait;
		}

		break;
    }
	case Wait:
		DisplayFrame();
		if (WaitCount++ < 100)
			Phase = Wait;
		else {
			Phase = Grow;
            if (sim_time + EPS >= 700) {
			    Stop();
		    }
        }
		break;
	}
	UseGroup(Phase);
}


// Move primordia centripetally as the receptale grows. Update age a.

group Grow:

P(x,a,num) :
{
	x += x * (rate-1.)*DT ;		// arc-length primordim coordinate x increases to preserve polar angle
    a += DT;

	produce P(x,a,num) ;
}

// Displace incipient primordium towards older neighbor

group Displace:

P(xl, al,nl) I(sl,pl) < P(x,a,num) > I(sr,pr) P(xr, ar,nr) :
{
	if (a<al && a < ar) {	// if P an incipient primordium
		if (al>=ar) {		// if left neighbor older than right, or break symmetry
			if (x < xl)		// circular arithmetic...
				xl -= C;	// ...enforced here
			x -=  k * ((x-xl)-Thd/2.0) * DT;	// displace w.r.t. the left neighbor
		}
		else { 				// right neighbor older than left
			if (x > xr)		// circular arithmetic...
				xr += C;	//...enforced here
			x += k * ((xr-x)-Thd/2.0) * DT;	// displace w.r.t. right neighbor
		}
	x = InInterval(x);		// map result to interval [0,c]
	}

	produce P(x,a,num) ;
}

// Insert new primordium if space between existing primordia is large enough;
// define the status and polarity of the adjacent intervals

group Insert:

P(xl,al,nl) < I(s,p) > P(xr,ar,nr) :
{
	if (xl>=xr)			// circular arithmetic...
		xr += C;		// enforced here
	if (xl+Thd <= xr) {	// if distance between primordia exceeds threshold
		++n;			// increment primordia count
		if (al >= ar) 	// test on which side is the older primordium 
			produce I(S,RIGHT) P(InInterval(xl+(xr-xl)*ratio), 1,n) I(L,LEFT) ;
		else 
			produce I(L,RIGHT) P(InInterval(xl+(xr-xl)*(1.-ratio)), 1,n) I(S,LEFT) ;
	}
	// produce I(s,p);
}

// Change interval status from S to L when a newer primordium is introduced

group AdjustStatus:
	
I(sl,pl) P(xl,al,nl) < I(s,p) > P(xr, ar,nr) I(sr,pr) :
{
	if (((p == RIGHT) && (pr == RIGHT)) || ((p == LEFT) && (pl == LEFT)))
			produce I(L,p);
}

// Do nothing: group necessary for syntactic reasons

group Wait:

Background : {produce Background(); }

interpretation:

group 0:

// Visualize primordium

static char strs[200][32];

P(x,a,num) :
{
	double r = R;
	if (a>250)
		a = 250;
    if (num < 200)
        sprintf(strs[num-1],"%d",num);
	produce SB SetWidth(0.085*1.5) SetColor(7) LineTo(-r*cos(x/R), -r*sin(x/R), 0) // draw "spoke"
		    SetColor(16+ int(a/2.5)) SB Up(90) f(1.0) Down(90)  Circle(0.65-0.085*0.5) SetColor(8) CircleB(0.65) EB EB;	// draw primordium
}

// Visualize interval 

P(xl,al,nl) < I(s,p) > P(xr,ar,nr) :
{
	// deal with the special case of a single primordium or crossing x==0 
	if (n == 1 || xl > xr) 
		xr += C;

	// define the increment for drawing intervals
	double dx = (xr-xl) / 50.0 ;

	double r1 = R-0.2;
	double r2 = R+0.2;

	V2d v1, v2, v3, v4;

	int color;

	// change "spoke" width from 0.2 to 0.15 for on-screen viewing.
	nproduce SB SetColor(color) MoveTo(-R*cos(xl/R), -R*sin(xl/R), 0) SetWidth(0.2);

	for(double x=xl; x<xr; x+=dx) {

		V2d v1 (-r1*cos(x/R), -r1*sin(x/R)); 
		V2d v2 (-r1*cos((x+dx)/R), -r1*sin((x+dx)/R)); 
		V2d v3 (-(r2)*cos((x+dx)/R), -(r2)*sin((x+dx)/R)); 
		V2d v4 (-(r2)*cos(x/R), -(r2)*sin(x/R)); 
	
		if (s==L)	// long intervals are gray
			color = 6; 
		else {
			if (p==LEFT) // short intervals have a color gradient, red points to the older neighbor
				color = 128 + (int) ((x-xl)/Thd*63);
			else
				color = 128 + (int) ((xr-x)/Thd*63);
		}
		nproduce SetColor(color);
		nproduce SB SP MoveTo2d(v1) PP MoveTo2d(v4) PP MoveTo2d(v3) PP MoveTo2d(v2) PP EP EB ;
	}
	produce EB ;
}

// Display confocal image at appropriate stage, position and orientation

#define SF 9. // scaling factor for background images

Background() :
{
	nproduce  Right(90) Up(180) SetColor(14);

    if (sim_time + EPS >= 621) // 34P
        produce MoveTo(-0.35,0.1,-0.5) Left(-1) CurrentTexture(7) Surface(0,SF);
    if (sim_time + EPS >= 535.) // 21P
		produce MoveTo(-0.2,0.1,-0.5) Left(-0) CurrentTexture(6) Surface (0, SF);
    if (sim_time + EPS >= 446.) // 13P
		produce MoveTo(-0.5,-0.3,-0.5) Left (0.5) CurrentTexture(5) Surface (0, SF);
    if (sim_time + EPS >= 358.) // 8P
		produce MoveTo(0.05,0.0,-0.5) Left(4) CurrentTexture(4) Surface (0, SF);
    if (sim_time + EPS >= 269.) // 5P
		produce MoveTo(0.,-0.1,-0.5) Left(-4) CurrentTexture(3) Surface (0, SF);
    if (sim_time + EPS >= 180.) // 3P
		produce MoveTo(0.1,-0.8,-0.5) Left(5) CurrentTexture(2) Surface (0, SF);
    if (sim_time + EPS >= 90.) // 2P
		produce MoveTo(0.0,-0.4,-0.5) Left(-2) CurrentTexture(1) Surface (0, SF);
	produce MoveTo(-0.25,-0.3,-0.5) Right(0) CurrentTexture(0) Surface (0, SF);
}

